/*
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; You may only use version 2 of the License,
	you have no option to use any other version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

        Metacity - (c) 2001 Havoc Pennington
        xfwm4    - (c) 2002 Olivier Fourdan

 */

#ifndef __MENU_H__
#define __MENU_H__

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <X11/Xlib.h>

typedef enum 
{
  MENU_OP_MAXIMIZE    = 1 << 0,
  MENU_OP_UNMAXIMIZE  = 1 << 1,
  MENU_OP_MINIMIZE    = 1 << 2,
  MENU_OP_UNMINIMIZE  = 1 << 3,
  MENU_OP_SHADE       = 1 << 4,
  MENU_OP_UNSHADE     = 1 << 5,
  MENU_OP_STICK       = 1 << 6,
  MENU_OP_UNSTICK     = 1 << 7,
  MENU_OP_MOVE        = 1 << 8,
  MENU_OP_RESIZE      = 1 << 9,
  MENU_OP_DELETE      = 1 << 10,
  MENU_OP_DESTROY     = 1 << 11,
  MENU_OP_WORKSPACES  = 1 << 12,
  MENU_OP_QUIT        = 1 << 13,
  MENU_OP_RESTART     = 1 << 14,
  MENU_OP_SWITCH      = 1 << 15,
  MENU_OP_OTHER       = 1 << 16
} MenuOp;

typedef struct _Menu Menu;
typedef struct _MenuItem MenuItem;
typedef struct _MenuData MenuData;

typedef void (* MenuFunc) (Menu    *menu,
			   MenuOp   op,
			   Window   client_xwindow,
			   gpointer menu_data,
			   gpointer item_data);

struct _MenuItem
{
  MenuOp      op;
  const char *stock_id;
  const char *label;
};

struct _MenuData
{
  Menu    *menu;
  MenuOp   op;
  Window   client_xwindow;
  gpointer data;
};

struct _Menu
{
  GtkWidget *menu;
  MenuFunc   func;
  gpointer   data;
  MenuOp     ops;
  MenuOp     insensitive;
};

Menu* menu_default (MenuOp ops, MenuOp insensitive, MenuFunc func, gpointer data);
Menu* menu_connect (Menu *menu);
GtkWidget*   menu_item_connect (GtkWidget *item, MenuData *item_data);
gboolean     menu_is_window (Window);
gboolean     menu_is_opened (void);
gboolean     menu_check_and_close (void);
void         menu_popup (Menu *menu, int root_x, int root_y, int button, guint32 timestamp);
void         menu_free (Menu *menu);

#endif /* __MENU_H__ */
